package c4.player;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import java.rmi.*;
import java.rmi.server.*;
import java.util.Vector;
import java.net.*;
import java.io.*;
import c4.shared.*;

// This creates the documentation
// javadoc -d C:\Will\Documentation -windowtitle Connect4 -doctitle
// Connect4 c4.server c4.player c4.shared
/**
 * The C4Player class represents the Connect 4 client program. It is responsible
 * for creating many of the game objects as well as the GUI and maintaining control.
 */
public class C4Player extends JFrame implements ActionListener,GameControllerListener,GameMovesListener
{
	/**
	 Constructs a C4Player object. All agregate objects are constructed in this method.
	 */
	public C4Player()
	{
		Container container=getContentPane();
		GridBagLayout gridBag=new GridBagLayout();
		GridBagConstraints gridConst=new GridBagConstraints();
		container.setLayout(gridBag);
		gridConst.anchor=GridBagConstraints.NORTHWEST;
		gridConst.fill=GridBagConstraints.HORIZONTAL;
		gridConst.weightx=0;//100;
		gridConst.weighty=0;//100;
		regPlayersPane=new RegPlayersPane();
		controlPane=new ControlPane(this);
		playerPane=new PlayerInfoPane();
		serverPane=new ServerInfoPane();
		statusPane=new StatusPane();
		opponentPane=new OpponentInfoPane();
		gamePane=new GamePane(this);
		JPanel panelForGame=new JPanel(); // not allowed to paint in a components region
   		TitledBorder gameBorder=new TitledBorder("Connect 4"); // so gamePane placed inside
   		panelForGame.setBorder(gameBorder); // panelForGame panel.
		panelForGame.add(gamePane);
		colourChooser=new ColourPane();
		soundEffects=new SoundEffects();
		soundEffectsPane=new SoundEffectsPane();
		game=new Game();
		playerQueue=new C4PlayerEventQueue(this);
		uiManager=new C4PlayerUIManager(serverPane,playerPane,statusPane,
		colourChooser,soundEffectsPane,controlPane,opponentPane);
		add(serverPane,gridConst,0,0,4,1);
		add(playerPane,gridConst,0,1,4,2);
		add(statusPane,gridConst,0,3,4,1);
		add(regPlayersPane,gridConst,0,4,1,3);//was 0,4,1,6
		add(colourChooser,gridConst,0,7,1,2);
		add(soundEffectsPane,gridConst,0,9,1,1);
		//add(gamePane,gridConst,1,4,2,6);
		add(panelForGame,gridConst,1,4,2,6);
		add(controlPane,gridConst,3,4,1,6);
		add(opponentPane,gridConst,0,10,4,1);

		addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				if (player!=null) {
					try {
						if (player.getStatus()==Player.PLAYING)
							handleEndEvent();
						if (registered)
							handleUnregisterEvent();
					} catch (Exception excep) {}
						// too late to do anything.
				}
				System.exit(0);
			}
		});

		try {
      	 	new GameControllerListenerThread(this).start();
      	 	new GameMovesListenerThread(this).start();
      	 } catch (Exception e) {
      	 	System.err.println("Can't start network listeners");
			String message=new String("Can't start network listeners.");
			String title=new String("Network Problem");
			showDialog(this,title,message);
      	 }

		timer=new Timer(1000,this);

		uiManager.setForUnregistered(); // sets UI status
	}

	private void add(Component c,GridBagConstraints g,int x,int y,int w,int h)
	{
		g.gridx=x;
		g.gridy=y;
		g.gridwidth=w;
		g.gridheight=h;
		getContentPane().add(c,g);
	}

	private void lookup(String url) throws Exception
	{
   	 	regPlayers=(RegPlayers)Naming.lookup("rmi://" + url + "/RegisteredPlayers");
   	 	gameController=(GameController)Naming.lookup("rmi://" + url + "/GameController");
   	 	gameMoves=(GameMoves)Naming.lookup("rmi://" + url + "/GameMoves");
   	 	if (DEBUG) System.out.println("Remote Object lookup..");
	}

	/**
	 * Processes a RegisterEvent, caused as a result of a player submitting a request
	 * to register with the connect 4 server. Method updates GUI through status pane,
	 * and button control panel and then issues a request to the server using RMI through
	 * RegPlayers remote object. RegisterEvent generated by clicking Un-register button
	 * on control panel.
	 * @see		RegisterEvent
	 * @see		RegPlayers
	 * @see		ControlPane
	 * @see		StatusPane
	 */
	public void handleRegisterEvent() throws Exception
	{
		if (registered) return;
		int validation=playerPane.checkEntry();
		if (validation!=PlayerInfoPane.OK) {
			String field=new String("Name");
			if (validation==PlayerInfoPane.LOCATION_FIELD)
			field=new String("Location");
			String message=new String("You haven't entered your " + field +" in.");
			String title=new String("Incorrect Data Entry");
			showDialog(this,title,message);
			if (DEBUG) System.out.println(message);
			return;
		}
		statusPane.setStatus("Registering...");
		lookup(serverPane.getServer());
		int reg=0;
   		if (registeredbefore) {
   			player.reset();
   			player.setName(playerPane.getPlayer().getName());
   			player.setLocation(playerPane.getPlayer().getLocation());
   			player.setStatus(Player.REGISTERED);
   			reg=regPlayers.registerPlayer(player,player.getID());
 		} else {
    			player=playerPane.getPlayer();
    			reg=regPlayers.registerPlayer(player);
			if (reg!=-1) {
   				player.setID(reg);
   				player.setStatus(Player.REGISTERED);
				Vector vector=regPlayers.getRegPlayers();
				regPlayersPane.setRegPlayers(vector,player);
   				regPlayersPane.init();
   				registeredbefore=true;
			}
		}
		if (reg==-1) {
			String title=new String("Registrar is full");
			String message=new String("Sorry, the C4Server is full. Please try later.");
			showDialog(this,title,message);
		} else {
			uiManager.setForRegistered();
   	 		registered=true;
		}
	}

	/**
	 * Processes an UnregisterEvent, caused as a result of a player submitting a request
	 * to un-register with the connect 4 server. Method updates GUI through status pane,
	 * and button control panel and then issues a request to the server using RMI through
	 * RegPlayers remote object. UnregisterEvent generated by clicking Un-register button
	 * on control panel.
	 * @see		UnregisterEvent
	 * @see		RegPlayers
	 * @see		ControlPane
	 * @see		StatusPane
	 */
	public void handleUnregisterEvent() throws Exception
	{
		if (!registered) return;
		statusPane.setStatus("Un-registering...");
		lookup(serverPane.getServer());
   		player.setStatus(Player.UNREGISTERED);
		regPlayers.unregisterPlayer(player);
		uiManager.setForUnregistered();
     	registered=false;
	}

	/**
	 * Processes a PlayEvent, caused as a result of a player submitting a request to play
	 * with an opponent registered with the server. Method checks to see if an opponent has been
	 * selected, if no opponent selected user is notified via a dialog. If an opponent is selected
	 * the is updated GUI through status pane and button control panel and a request to the server
	 * is issued using RMI through GameController remote object. PlayEvent generated by clicking
	 * Play button on control panel.
	 * @see		PlayEvent
	 * @see		GameController
	 * @see		ControlPane
	 * @see		StatusPane
	 * @see		showDialog
	 */
	public void handlePlayEvent() throws Exception
	{
		Player opponent=regPlayersPane.getSelectedOpponent();
		if (opponent==null) {
			String message=new String("You haven't selected an opponent");
			String title=new String("No Opponent");
			showDialog(this,title,message);
			if (DEBUG) System.out.println(message);
			return;
		}
		uiManager.setForRequesting();
		player.setStatus(Player.REQUESTING);
   		regPlayers.updatePlayer(player);
		game.setGame(player,opponent);
		if (opponent.getID()==-1) { // play the computer
			if (DEBUG) System.out.println("Your requesting the Computer");
			Game computerGame=new Game();
			computerGame.setGame(opponent,player);
			playingTheComputer=true;
			playerQueue.postEvent(new GameAcceptedEvent(this,computerGame));
		} else {
			boolean	bool=gameController.requestGame(game);
			timer.start();
			delays=0;
			if (DEBUG) System.out.println("You've selected to play " + opponent.getName());
			if (!bool) {
				String title=new String("Games Limit");
				String message=new String("Sorry, the C4Server has reached it maximum permitted games limit. Please try again later.");
				showDialog(this,title,message);
				player.setStatus(Player.REGISTERED);
      			regPlayers.updatePlayer(player);
				uiManager.setForRegistered();
			}
		}
	}

	/**
	 * Processes an EndEvent, caused as a result of a player submitting a request to terminate an
	 * active game with an opponent. Method updates GUI through status pane and button control panel
	 * and then issues a request to the server to prematurely end the game using RMI through
	 * GameController remote object. EndEvent generated by clicking Finish button on control panel.
	 * @see		EndEvent
	 * @see		GameController
	 * @see		ControlPane
	 * @see		StatusPane
	 */
	public void handleEndEvent() throws Exception
	{
		timer.stop();
		delays=0;
		game.setGame(player,game.getOpponent());
		gameController.endGame(game);
		player.setStatus(Player.REGISTERED);
		regPlayers.updatePlayer(player);
		gamePane.setMyMove(false);
		//gamePane.stop(); leave commented out for moment
		uiManager.setForRegistered();
		playingTheComputer=false;
	}

	/**
	 * Processes a GameRequestEvent, caused as a result of a player receiving a game request from an opponent
	 * Method changes player status to requesting and displays a dialog asking player to accept/decline game
	 * challenge. The button control control panel and status panel are updated accordingly.
	 * @param	evt	the GameRequestEvent
	 * @see		GameRequestEvent
	 * @see		ControlPane
	 * @see		StatusPane
	 * @see		showRequestDialog
	 */
	public void handleGameRequestEvent(GameRequestEvent evt) throws Exception
	{
		timeElapsed=false;
		timer.start();
		delays=0;
		if (getState()==Frame.ICONIFIED) // deiconifies the window if necessary
			setState(Frame.NORMAL);
		player.setStatus(Player.RECEIVED_REQUEST);
		statusPane.setStatus("Game request received...");
		Game gameRequest=evt.getGame();
		//controlPane.setDisabled(); // don't need to do this, dialog is modal
		Player opponent=gameRequest.getPlayer();
		String message=new String(opponent.getName() + " would like to play you, do you accept?");
		String title=new String("Game Request");
		game.setGame(player,gameRequest.getPlayer());
		try {
			showRequestDialog(this,title,message);
		} catch (Exception e) {
			throw e;
		}
	}

	/**
	 * Processes a GameAcceptedEvent, caused as a result of a player accepting a game challenge.
	 * Method updates the player's object status to playing and sends updated object to server
	 * though remote RegPlayers object. The GUI status panel is updated as is the button control panel.
	 * The game panel is initialised to start the game and the player is set to take the first move.
	 * This event is generated by clicking the 'Yes' button on the showRequestDialog.
	 * on control panel.
	 * @param	evt	the GameAcceptedEvent
	 * @see		PlayEvent
	 * @see		RegPlayers
	 * @see		ControlPane
	 * @see		StatusPane
	 * @see		showRequestDialog
	 * @see		GamePane
	 */
	public void handleGameAcceptedEvent(GameAcceptedEvent evt) throws Exception
	{
		timer.start();
		delays=0;
		Game gameAccepted=evt.getGame();
		uiManager.setForPlaying(game.getOpponent(),"Waiting for opponent to move...");
		soundOn=soundEffectsPane.isSoundOn();
		if (soundOn)
			soundEffects.play(SoundEffects.START);
		game.setGame(player,gameAccepted.getPlayer());
		player.setStatus(Player.PLAYING);
		regPlayers.updatePlayer(player);
		String message=new String(game.getOpponent().getName() + " has accepted your challenge, GAME ON!");
		String title=new String("Game Accepted");
		showDialog(this,title,message);
		if (DEBUG) System.out.println(message);
		gamePane.setColour(colourChooser.getColour());
		gamePane.setMyMove(false);
		gamePane.reset(player);
		//gamePane.start(); // now in GamePane constructor
		if (game.getOpponent().getID()==-1) { // opponent is the computer
			int result=gamePane.computerMove();
			if (soundOn)
				soundEffects.play(SoundEffects.MOVE);
		}

	}

	/**
	 * Processes a GameDeclinedEvent, caused as a result of a player declining a game challenge.
	 * Method updates the player's object status to playing and sends updated object to server
	 * though remote RegPlayers object. The GUI status panel is updated as is the button control panel.
	 * This event is generated by clicking the 'No' button on the showRequestDialog.
	 * on control panel.
	 * @param	evt	the GameDeclinedEvent
	 * @see		PlayEvent
	 * @see		RegPlayers
	 * @see		ControlPane
	 * @see		StatusPane
	 * @see		showRequestDialog
	 */
	public void handleGameDeclinedEvent(GameDeclinedEvent evt) throws Exception
	{
		timer.stop();
		Game gameDeclined=evt.getGame();
		uiManager.setForRegistered();
		game=gameDeclined;
		player.setStatus(Player.REGISTERED);
		regPlayers.updatePlayer(player);
		String message=new String("Sorry, but " + game.getPlayer().getName() + " has rejected your challenge");
		String title=new String("Game Rejected");
		showDialog(this,title,message);
		if (DEBUG) System.out.println(message);
	}

	/**
	 * Processes a GameEndedEvent, caused as a result of an opponent terminating a game permaturely.
	 * Method updates the player's object status to playing and sends updated object to server
	 * though remote RegPlayers object. The GUI status panel is updated as is the button control panel.
	 * A dialog is displayed via showDialog method and game panel is notify to stop diplaying moves.
	 * @param	evt	the GameEndedEvent
	 * @see		GameEndedEvent
	 * @see		RegPlayers
	 * @see		ControlPane
	 * @see		StatusPane
	 * @see		showDialog
	 */
	public void handleGameEndedEvent(GameEndedEvent evt) throws Exception
	{
		timer.stop();
		Game gameEnded=evt.getGame();
		uiManager.setForRegistered();
		player.setStatus(Player.REGISTERED);
		regPlayers.updatePlayer(player);
		String message=new String("Sorry, but " + gameEnded.getPlayer().getName() +
				" has prematurely ended the Game");
		String title=new String("Game Over");
		showDialog(this,title,message);
		if (DEBUG) System.out.println(message);
		gamePane.setMyMove(false);
		//gamePane.stop();
		playingTheComputer=false;

	}

	/**
	 * Processes a BusyEvent, used to prevent deadlock caused by processing and network delays
	 * which can result in two players simultaneously requesting each other. The method issues a
	 * response back to the server so that the opponent challenger can the be notified that the player
	 * is busy requesting another player. The notification is propagated using the GameController remote
	 * object.
	 * @param	evt	the BusyEvent
	 * @see		BusyEvent
	 * @see		GameController
	 */
	public void handleBusyEvent(BusyEvent evt) throws Exception
	{
		Game gameBusy=evt.getGame();
		//uiManager.setForRegistered();
		gameBusy.setGame(player,gameBusy.getPlayer());
		if (DEBUG) System.out.println("Player " + player.getName() + " Opponent " + gameBusy.getPlayer().getName());
		gameController.playerIsBusy(gameBusy);
	}

	/**
	 * Processes an OpponentBusyEvent, this is event is caused by an opponent who was requesting to
	 * play when he/she received and a request. This is used to prevent a dealock condition. The event is
	 * caused by a BusyEvent. Method updates the GUI through the status panel and button control panel
	 * changes the player status back to registered an sends the updated player object to the server. A
	 * dialog is displayed to reflect that the requested opponent was busy.
	 * @param	evt	the OpponentBusyEvent
	 * @see		OpponentBusyEvent
	 * @see		BusyEvent
	 * @see		RegPlayers
	 * @see		ControlPane
	 * @see		StatusPane
	 * @see		showDialog
	 */
	public void handleOpponentBusyEvent(OpponentBusyEvent evt) throws Exception
	{
		Game gameOpponBusy=evt.getGame();
		uiManager.setForRegistered();
		player.setStatus(Player.REGISTERED);
		String message=new String("Sorry, but " + gameOpponBusy.getPlayer().getName() + " is busy, please try again later");
		String title=new String("Game Over");
		showDialog(this,title,message);
		if (DEBUG) System.out.println(message);
		regPlayers.updatePlayer(player);
	}

	/**
	 * Processes a ReceivedMoveEvent, this event is generated when a move is received via the
	 * gameMovesListener interface. Method deals with moves that are generated by either the
	 * player or the opponent. The status panel is updated accordingly. The move is transmittted
	 * to the game panel. If the move is from the player it is also sent to the server for
	 * announcement to the opponent through the GameMoves remote object.
	 * @param	evt	the ReceivedMoveEvent
	 * @see		ReceivedMoveEvent
	 * @see		GameMoves
	 * @see		StatusPane
	 */
	public void handleReceivedMoveEvent(ReceivedMoveEvent evt) throws Exception
	{
		int result=0;
		Move move=evt.getMove();
		int id=move.getPlayer().getID();
		if (DEBUG) System.out.println("Moves from " + move.getPlayer().getName());
		if (playingTheComputer) {
			if (id==player.getID() && gamePane.isMyMove()) {
				result=gamePane.move(move);
				if (soundOn)
					soundEffects.play(SoundEffects.MOVE);
				statusPane.setStatus("Waiting for opponent to move...");
				if (result==0) {
					result=gamePane.computerMove();
					if (soundOn)
						soundEffects.play(SoundEffects.MOVE);
				}
			}
		} else if (id==game.getOpponent().getID() && !gamePane.isMyMove()) { // opponents move
		//if (id==game.getOpponent().getID() && !gamePane.isMyMove()) { // opponents move
			result=gamePane.move(move);
			if (soundOn)
				soundEffects.play(SoundEffects.MOVE);
			statusPane.setStatus("It's your move");
		} else if (id==player.getID() && gamePane.isMyMove()) { // my move
			gameMoves.move(move); // send move to server
			result=gamePane.move(move);
			if (soundOn)
				soundEffects.play(SoundEffects.MOVE);
			statusPane.setStatus("Waiting for opponent to move...");
		}
		if (result!=0) {
			timer.stop();
			playerQueue.postEvent(new ResultEvent(this,result));
		} else
			timer.restart();
			delays=0;
	}

	/**
	 * Processes a ResultEventEvent, this event is generated when a game is either won, drawn or lost.
	 * The status panel, button control panel, player and player information panel are all updated. A
	 * dialog  is then diplayed showing the result of the game. The server is contacted and sent the
	 * updated player object an also informed that the game has finished, this is done through the
	 * RegPlayers and GameController remote obkects.
	 * @param	evt	the ResultEvent
	 * @see		GameController
	 * @see		RegPlayers
	 * @see		StatusPane
	 * @see		ControllPane
	 * @see		showDialog
	 */
	public void handleResultEvent(ResultEvent evt) throws Exception
	{
		timer.stop();
		String message=null;
		int result=evt.getResult();
		if (result==ResultEvent.WON) {
			if (soundOn)
				soundEffects.play(SoundEffects.WON);
			message=new String("You've WON!");
			player.incWin();
		} else if (result==ResultEvent.DRAWN) {
			message=new String("You've DRAWN!");
			player.incDrawn();
		} else if (result==ResultEvent.LOST) {
			if (soundOn)
				soundEffects.play(SoundEffects.LOST);
			message=new String("You've LOST!");
			player.incLost();
		}
		gamePane.setMyMove(false);
		player.incPlayed();
		playerPane.updatePlayer(player);
		String title=new String("Game Result");
		showDialog(this,title,message);
		uiManager.setForRegistered();
		player.setStatus(Player.REGISTERED);
		regPlayers.updatePlayer(player);
		gameController.finishGame(game);
		playingTheComputer=false;
	}


	/**
	 * Processes a PlayerStats event which displays a selected players game statistics. The players
	 * details are displayed in a message dialog. If no player is selected an appropriate dialog is
	 * displayed.
	 * @param	evt the PlayerStatsEvent
	 * @see		showDialog
	 */
	public void handlePlayerStatsEvent(PlayerStatsEvent evt)
	{
		Player playerStats=regPlayersPane.getSelectedOpponent();
		if (playerStats==null) {
			String message=new String("You haven't selected an opponent");
			String title=new String("No Opponent");
			showDialog(this,title,message);
			if (DEBUG) System.out.println(message);
			return;
		}
		String title=new String("Player Statistics");
		JPanel message=new JPanel();
		message.setLayout(new GridLayout(6,2));
		message.add(new Label("Name"));
		message.add(new Label(playerStats.getName()));
		message.add(new Label("Location"));
		message.add(new Label(playerStats.getLocation()));
		message.add(new Label("Played"));
		message.add(new Label("" + playerStats.getPlayed()));
		message.add(new Label("Won"));
		message.add(new Label("" + playerStats.getWin()));
		message.add(new Label("Drawn"));
		message.add(new Label("" + playerStats.getDrawn()));
		message.add(new Label("Lost"));
		message.add(new Label("" + playerStats.getLost()));
		showDialog(this,title,message);
	}

	public void handleTimeElapsedEvent(TimeElapsedEvent evt) throws Exception
	{
		String message=null;
		String title=new String("Game Time-out");
		int status=player.getStatus();
		if (status==Player.PLAYING && !gamePane.isMyMove()) {
			if (delays<MOVE_DELAYS+6) {
				if ((MOVE_DELAYS-delays+DELAY_FLOAT)<=10) {
					message=new String("" + (MOVE_DELAYS-delays+DELAY_FLOAT));
					statusPane.setStatus("Waiting for opponent to move..." + message);
				}
				++delays;
			} else {
				player.setStatus(Player.REGISTERED);
   				regPlayers.updatePlayer(player);
				timer.stop();
				playerQueue.postEvent(new EndEvent(this));
				uiManager.setForRegistered();
				message=new String("No move received from opponent within time-out period.");
				if (soundOn)
					soundEffects.play(SoundEffects.ALERT);
				showDialog(this,title,message);
			}
		} else if (status==Player.PLAYING && gamePane.isMyMove()) {
			if (delays<MOVE_DELAYS) {
				if ((MOVE_DELAYS-delays)<=10){
					message=new String("" + (MOVE_DELAYS-delays));
					if (soundOn)
						soundEffects.play(SoundEffects.ALERT);
					statusPane.setStatus("It's your move..." + message);
				}
				++delays;
			} else {
				player.setStatus(Player.REGISTERED);
   				regPlayers.updatePlayer(player);
				timer.stop();
				gamePane.setMyMove(false);
				uiManager.setForRegistered();
				playingTheComputer=false;
				message=new String("Time out has occurred. GAME OVER.");
				if (soundOn)
					soundEffects.play(SoundEffects.ALERT);
				showDialog(this,title,message);
			}
		} else if (status==Player.REQUESTING){
			if (delays<REQUEST_DELAYS+DELAY_FLOAT) {
				++delays;
			} else {
				timer.stop();
				player.setStatus(Player.REGISTERED);
   				regPlayers.updatePlayer(player);
				uiManager.setForRegistered();
				message=new String("No response received from opponent.");
				if (soundOn)
					soundEffects.play(SoundEffects.ALERT);
				showDialog(this,title,message);
			}
		} else if (status==Player.RECEIVED_REQUEST){
			if (delays<REQUEST_DELAYS) {
				++delays;
			} else {
				timeElapsed=true;
				timer.stop();
				player.setStatus(Player.REGISTERED);
   				regPlayers.updatePlayer(player);
				uiManager.setForRegistered();
			}
		}
	}


	/**
	 * Displays a Message dialog with and 'OK' button. Parameters passed to methodDialog set the
	 * title and the message as well as the container to which the display relative is to be
	 * relative to.
	 * @param	dContainer the container to which the dialog display is to be relative to.
	 * @param	dTitle the title of the dialog
	 * @param	dMessage the message object to be displayed
	 */
	public void showDialog(C4Player dContainer,String dTitle,Object dMessage)
	{
		final C4Player c4Player=dContainer;
		final String title=dTitle;
		final Object message=dMessage;
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				JOptionPane.showMessageDialog(c4Player,message,title,JOptionPane.PLAIN_MESSAGE);
				if (soundEffectsPane.isSoundOn())
					soundEffects.play(SoundEffects.CLICK);
			}
		});
	}

	/**
	 * Displays an Option dialog which asks the player if the would like to accept a game challenge.
	 * The dialog displays both 'Yes' and 'No' buttons. Parameters passed to methodDialog set
	 * the title and the message as well as the container to which the display relative is to be
	 * relative to. The GUI is updated in a fashion which reflects the players response to the dialog.
	 * If the player clicks 'Yes' the player object is updated to the playing status and sent to the
	 * server via the RegPlayers remote object. The GameController remote object is used to sned the
	 * result of the request back to the server, which then informs the challenging opponent.
	 * @param	dContainer the container to which the dialog display is to be relative to.
	 * @param	dTitle the title of the dialog
	 * @param	dMessage the message to be displayed
	 * @see		Player
	 * @see		GameController
	 * @see		RegPlayers
	 * @see		StatusPane
	 * @see		ControllPane
	 */
	public void showRequestDialog(C4Player dContainer,String dTitle,String dMessage) throws Exception
	{
		final C4Player c4Player=dContainer;
		final String title=dTitle;
		final String message=dMessage;
		if (soundEffectsPane.isSoundOn())
			soundEffects.play(SoundEffects.ALERT);
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				final int answer=JOptionPane.showOptionDialog(c4Player,message,title,JOptionPane.YES_NO_OPTION,JOptionPane.QUESTION_MESSAGE,null,null,null);
				if (soundEffectsPane.isSoundOn())
					soundEffects.play(SoundEffects.CLICK);
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						//game.setGame(player,gameRequest.getPlayer());
						if (DEBUG) System.out.println(message);
						if (timeElapsed) {
							String lateMessage=new String("Sorry, but your response was to late");
							String lateTitle=new String("Response to late");
							showDialog(c4Player,lateTitle,lateMessage);
							if (soundEffectsPane.isSoundOn())
								soundEffects.play(SoundEffects.CLICK);
							return;
						}
						if (answer==0) {
							if (soundEffectsPane.isSoundOn())
								soundEffects.play(SoundEffects.START);
							player.setStatus(Player.PLAYING);
							try {
								regPlayers.updatePlayer(player);
								gameController.acceptGame(game);
							} catch (RemoteException e) {
								System.err.println("Remote Exception " + e);
								String netMessage=new String("A network error has been detected.");
								String netTitle=new String("Network Problem");
								showDialog(c4Player,netTitle,netMessage);
							}
							if (DEBUG) System.out.println("Accepted");
							uiManager.setForPlaying(game.getOpponent(),"It's your move");
							soundOn=soundEffectsPane.isSoundOn();
							gamePane.setColour(colourChooser.getColour());
							gamePane.setMyMove(true);
							gamePane.reset(player);
							//gamePane.start(); // now in GamePane constructor
							timer.restart();
							delays=0;
						} else {
							timer.stop();
							try {
								gameController.declineGame(game);
							} catch (RemoteException e) {
								System.err.println("Remote Exception " + e);
								String netMessage=new String("A network error has been detected.");
								String netTitle=new String("Network Problem");
								showDialog(c4Player,netTitle,netMessage);
							}
							player.setStatus(Player.REGISTERED);
							if (DEBUG) System.out.println("Declined");
							uiManager.setForRegistered();
						}
					}
				});
			}
		});
	}

	/**
	 * Handles all of the various C4Player events. Calls upon handleTYPEEvent methods in C4Player class
	 * @param	event and AWTEvent
	 */
	public void processEvent(AWTEvent event)
	{
		String message=null;
		String title=new String("Network Problem");
	  	try {
			if (event instanceof RegisterEvent) {
				if (DEBUG) System.out.println("processing RegisterEvent");
				handleRegisterEvent();
			} else if (event instanceof UnregisterEvent) {
				if (DEBUG) System.out.println("processing UnregisterEvent");
				handleUnregisterEvent();
			} else if (event instanceof PlayEvent) {
				if (DEBUG) System.out.println("processing PlayEvent");
				handlePlayEvent();
			} else if (event instanceof EndEvent) {
				if (DEBUG) System.out.println("processing EndEvent");
				handleEndEvent();
			} else if (event instanceof GameRequestEvent) {
				if (DEBUG) System.out.println("processing GameRequestEvent");
				handleGameRequestEvent((GameRequestEvent)event);
			} else if (event instanceof GameAcceptedEvent) {
				if (DEBUG) System.out.println("processing GameAcceptedEvent");
				handleGameAcceptedEvent((GameAcceptedEvent)event);
			} else if (event instanceof GameDeclinedEvent) {
				if (DEBUG) System.out.println("processing GameDeclinedEvent");
				handleGameDeclinedEvent((GameDeclinedEvent)event);
			} else if (event instanceof GameEndedEvent) {
				if (DEBUG) System.out.println("processing GameEndedEvent");
				handleGameEndedEvent((GameEndedEvent)event);
			} else if (event instanceof BusyEvent) {
				if (DEBUG) System.out.println("processing BusyEvent");
				handleBusyEvent((BusyEvent)event);
			} else if (event instanceof OpponentBusyEvent) {
				if (DEBUG) System.out.println("processing OpponentBusyEvent");
				handleOpponentBusyEvent((OpponentBusyEvent)event);
			} else if (event instanceof ReceivedMoveEvent) {
				if (DEBUG) System.out.println("processing ReceivedMoveEvent");
				handleReceivedMoveEvent((ReceivedMoveEvent)event);
			} else if (event instanceof ResultEvent) {
				if (DEBUG) System.out.println("processing ResultEvent");
				handleResultEvent((ResultEvent)event);
			} else if (event instanceof PlayerStatsEvent) {
				if (DEBUG) System.out.println("processing PlayerStatsEvent");
				handlePlayerStatsEvent((PlayerStatsEvent)event);
			} else if (event instanceof TimeElapsedEvent) {
				//if (DEBUG) System.out.println("processing TimeElapsedEvent");
				handleTimeElapsedEvent((TimeElapsedEvent)event);
			} else {
				super.processEvent(event);
			}
		} catch (java.rmi.NotBoundException exc) {
		    System.err.println("Couldn't find C4Server running on host.");
		    System.err.println("RMI seems to be running fine.");
			message=new String("Couldn't find C4Server running on host.\nRMI seems to be running fine.");
			showDialog(this,title,message);
		} catch (java.rmi.UnmarshalException exc) {
			System.err.println("Unmarshal exception on host.");
			System.err.println("Try recompiling everything and starting over?");
			message=new String("Unmarshal exception on host.\nTry recompiling everything and starting over?");
			showDialog(this,title,message);
		} catch (java.rmi.ConnectException exc) {
		    System.err.println("RMI isn't running on host.");
		 	System.err.println("Or maybe it is, but it started after this program.");
			message=new String("RMI isn't running on host.\nOr maybe it is, but it started after this program.");
			showDialog(this,title,message);
		} catch (java.rmi.UnknownHostException exc) {
		    System.err.println("Unknown host");
			message=new String("Unknown host.");
			showDialog(this,title,message);
		} catch (java.rmi.ConnectIOException exc) {
			System.err.println("Couldn't get to host.");
			message=new String("Couldn't get to host.");
			showDialog(this,title,message);
		} catch (Exception exc) {
			System.err.println("Unexpected exception when trying to find registrar.");
			System.err.println("Exception type: " + exc.toString());
			exc.printStackTrace();
			System.exit(-1);
		}
	}

	/**
	 * Implemented from GameMovesListener remote interface. Receives notifications of moves.
	 * @param	move the Move
	 * @see		GameMovesListener
	 */
	public synchronized void updateMoves(Move move)
	{
		if (DEBUG) System.out.println("Received a move");
		if (player==null)
			return;
		else if (player.getStatus()!=Player.PLAYING)
			return;
		else if (!move.isMoveForMe(game.getPlayer(),game.getOpponent()))
			return;
		else {
			playerQueue.postEvent(new ReceivedMoveEvent(this,move));
		}
	}

	/**
	 * Implemented from GameControllerListener remote interface. Receives notifications of game accepts,
	 * declines and ends.
	 * @param	game the Game
	 * @see		GameControllerListener
	 */
	public synchronized void updateGames(Game game)
	{
   		if (!registeredbefore || player==null) return;
		if (!game.isGameForMe(player)) return;
		if (DEBUG) System.out.println("Received game controller message");
		int gameStatus=game.getStatus();
		int playerStatus=player.getStatus();
		//if (playerStatus==Player.RECEIVED_REQUEST) {
		//	playerQueue.postEvent(new BusyEvent(this,game));
		//	return;
		//}
		//if (gameStatus==Game.REQUEST) {
		//} else if (gameStatus==Game.REQUEST) {
		if (gameStatus==Game.REQUEST) {
			if (playerStatus==Player.REGISTERED)
				playerQueue.postEvent(new GameRequestEvent(this,game));
			else
				playerQueue.postEvent(new BusyEvent(this,game));
		} else if (gameStatus==Game.ACCEPT) {
			if (playerStatus==Player.REQUESTING)
				playerQueue.postEvent(new GameAcceptedEvent(this,game));
			else
				playerQueue.postEvent(new BusyEvent(this,game));
		} else if (gameStatus==Game.DECLINE) {
			if (playerStatus==Player.REQUESTING)
				playerQueue.postEvent(new GameDeclinedEvent(this,game));
			else
				playerQueue.postEvent(new BusyEvent(this,game));
		} else if (gameStatus==Game.END) {
			if (playerStatus==Player.PLAYING)
				playerQueue.postEvent(new GameEndedEvent(this,game));
			else
				playerQueue.postEvent(new BusyEvent(this,game));
		} else if (gameStatus==Game.BUSY && playerStatus==Player.REQUESTING) {
			playerQueue.postEvent(new OpponentBusyEvent(this,game));
		}
	}

	/**
	 * Implemented from ActionListener. Receives events generated by the button control panel.
	 * @param	e the ActionEvent
	 * @see		ActionListener
	 */
 	public void actionPerformed(ActionEvent e)
    {
		if (e.getSource()==timer) {
			//timer.restart();
			//if (DEBUG) System.out.println("timer");
			playerQueue.postEvent(new TimeElapsedEvent(this));
			return;
		}
		if (soundEffectsPane.isSoundOn())
			soundEffects.play(SoundEffects.CLICK);
		String command=e.getActionCommand();
		if (DEBUG) System.out.println(command);
		if (command.equals(this.REGISTER))
			playerQueue.postEvent(new RegisterEvent(this));
		if (command.equals(this.UNREGISTER))
			playerQueue.postEvent(new UnregisterEvent(this));
		if (command.equals(this.PLAY))
			playerQueue.postEvent(new PlayEvent(this));
		if (command.equals(this.END))
			playerQueue.postEvent(new EndEvent(this));
		if (command.equals(this.PLAYER))
			playerQueue.postEvent(new PlayerStatsEvent(this));
		if (command.equals(this.EXIT)) {
			if (soundEffectsPane.isSoundOn())
					soundEffects.play(SoundEffects.CLICK);
			if (player!=null) {
				if (player.getStatus()==Player.PLAYING)
					playerQueue.postEvent(new EndEvent(this));
				if (registered)
					playerQueue.postEvent(new UnregisterEvent(this));
			}
			System.exit(0);
		}
	}

	/**
	 * Creates and displays the C4Player object.
	 */
	public static void main(String[] args)
   	{
   		C4Player c4Player=new C4Player();
		//c4Player.setSize(500,500);
		c4Player.setTitle("C4Player");
		c4Player.pack();
		c4Player.setVisible(true);
   	}

	private Timer timer;
	private boolean timeElapsed=false;
	private boolean playingTheComputer=false;

	private SoundEffects soundEffects;
	private SoundEffectsPane soundEffectsPane;
	private ColourPane colourChooser;
	private RegPlayers regPlayers;
	private Player player=null;
	private Game game;
	private Timer gameTimer;
	private boolean registered=false;
	private boolean registeredbefore=false;
	private C4PlayerEventQueue playerQueue;
	private RegPlayersPane regPlayersPane;
	private GameController gameController;
	private GameMoves gameMoves;
	private ControlPane controlPane;
	private PlayerInfoPane playerPane;
	private ServerInfoPane serverPane;
	private StatusPane statusPane;
	private OpponentInfoPane opponentPane;
	private GamePane gamePane;

	private C4PlayerUIManager uiManager;

	private final static int REQUEST_DELAYS=30;
	private final static int MOVE_DELAYS=60;
	private final static int DELAY_FLOAT=10;
	private int delays=0;

	private boolean soundOn=true;
	//private Container container;
	private boolean DEBUG=true;
	/** Used to determine which control pane button was pressed.*/
	public final static String REGISTER=new String("register");
	/** Used to determine which control pane button was pressed.*/
	public final static String UNREGISTER=new String("unregister");
	/** Used to determine which control pane button was pressed.*/
	public final static String PLAY=new String("play");
	/** Used to determine which control pane button was pressed.*/
	public final static String END=new String("finish");
	/** Used to determine which control pane button was pressed.*/
	public final static String PLAYER=new String("player");
	/** Used to determine which control pane button was pressed.*/
	public final static String EXIT=new String("exit");
}