package c4.player;

import java.util.Vector;
import javax.swing.border.*;
import javax.swing.*;
import c4.shared.*;

/**
 * UI component which displays a list of players available to play. This class implements the
 * RegPlayersListener interface which provides player updates which arrive from a RegPlayerListenerThread
 * object. Players can be selected from this component which can then be used to make game requests.
 * @see		RegPlayersListener
 * @see		RegPlayerListenerThread
 */
public class RegPlayersPane extends JPanel implements RegPlayersListener
{
	/**
	 * Creates the RegPlayersListener object
	 */
	public RegPlayersPane()
	{
		model=new DefaultListModel();
		players=new JList(model);
		players.setFixedCellWidth(150);
		players.setFixedCellHeight(14);
		selection=new DefaultListSelectionModel();
		selection.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		selection.setLeadAnchorNotificationEnabled(false);
		players.setSelectionModel(selection);
		JScrollPane playerPane=new JScrollPane(players);
		playerVector=new Vector(10,1);
		add(playerPane);
   		TitledBorder regPlayerBorder=new TitledBorder("Registered Players");
   		setBorder(regPlayerBorder);
	}

	/**
	 * Initialises the component by starting the RegPlayerListenerThread object.
	 */
	public void init()
	{
		try {
      	 		new RegPlayersListenerThread(this).start();
      	 	} catch (Exception e) {
      	 		System.out.println("Can't start listener");
      	 	}
	}

	/**
	 * Initialises the component with a list of registered players.
	 * @param vector the Vector of registered players.
	 * @param me the player, not listed in the display (you can't play yourself).
	 */
	public synchronized void setRegPlayers(Vector vector,Player me)
	{
		myPlayer=me;
		Player computer=new Player("Computer","Right here!");
		computer.setID(-1);
		model.addElement("Computer");
		playerVector.addElement(computer);
		if (DEBUG) System.out.println("In setRegPlayers");
		int myID=myPlayer.getID();
		for(int i=0;i<vector.size();++i) {
			Player player=(Player)vector.elementAt(i);
			if (myID!=player.getID() && player.getStatus()!=Player.PLAYING) {
				model.addElement(player.getName());
				playerVector.addElement(player);
			}
			if (DEBUG) System.out.println("playerVector size " + playerVector.size());
		}
	}

	/**
	 * Add a player to the display.
	 * @param	player the player to be added to the display.
	 */
	public synchronized void addPlayer(Player player)
	{
		if (DEBUG) System.out.println("In addPlayer");
		playerVector.addElement(player);
		model.addElement(player.getName());
	}

	/**
	 * Updates the players list. Removes any playing players and adds players with a registered status.
	 * @param	player the player with which the display is to be updated.
	 */
	public synchronized void updatePlayer(Player player)
	{
		if (DEBUG) System.out.println("In updatePlayer, updating:" + player.getName());
		int i=findPlayerPosition(player);
		if (i!=-1) {
			Player p=(Player)playerVector.elementAt(i);
			p.updatePlayer(player);
			if (p.getStatus()==Player.PLAYING) {
				if (DEBUG) System.out.println("Removing " + player.getName());
				playerVector.removeElementAt(i);
				model.removeElementAt(i);
			}
		} else if (player.getStatus()==Player.REGISTERED) {
			if (DEBUG) System.out.println("Adding " + player.getName());
			playerVector.addElement(player);
			model.addElement(player.getName());
		}
	}

	/**
	 * Removes a player from the display.
	 * @param	player the player to be removed.
	 */
	public synchronized void removePlayer(Player player)
	{
		if (DEBUG) System.out.println("In removePlayer");
		int i=findPlayerPosition(player);
		if (i!=-1) {
			playerVector.removeElementAt(i);
			model.removeElementAt(i);
		}
	}

	/**
	 * Returns the displays selected player.
	 * @return	player the selected player. If no player is selected return value is null.
	 */
	public Player getSelectedOpponent()
	{
		if (!players.isSelectionEmpty()) {
			int selected=players.getSelectedIndex();
			return (Player)playerVector.elementAt(selected);
		} else {
			if (DEBUG) System.out.println("Selection empty");
			return null;
		}
	}

	/**
	 * RegPlayersListener implemented method, receives player objects to update the display.
	 * @param	player a player object to be updated in accordance with the players udpate type.
	 * @see		RegPlayersListenerThread
	 * @see		RegPlayersListener
	 * @see		Player
	 */
   	public synchronized void updateRegPlayers(Player player)
   	{
   		if (myPlayer.getID()==player.getID()) return;
   		int change=player.getUpdate();
		if (DEBUG) System.out.println("Id " + player.getID());
		if (change==Player.ADD_NEW_PLAYER) addPlayer(player);
		if (change==Player.UPDATE_PLAYER) updatePlayer(player);
		if (change==Player.REMOVE_PLAYER) removePlayer(player);
    }

	private int findPlayerPosition(Player player)
	{
		int i;
		int id=player.getID();
		int size=playerVector.size();
		boolean found=false;
		for (i=0;i<size;i++) {
			if (DEBUG) System.out.println("looking at element " + i + " of " + size);
			Player p=(Player)playerVector.elementAt(i);
			if (p.getID()==id) {
				if (DEBUG) System.out.println("found at " + i);
				found=true;
				return i;
			}
		}
		if (DEBUG && found) System.out.println("Found player position at element " + i);
		return -1;
	}

	private boolean DEBUG=false;
	private Player myPlayer;
	private JList players;
	private Vector playerVector;
   	private RegPlayers regPlayers;
	private DefaultListModel model;
	private DefaultListSelectionModel selection;
}

