package c4.server;

import java.util.Vector;
import java.rmi.*;
import java.rmi.server.*;
import c4.shared.*;

/**
 * Used to manage games between players and the server. Implements the GameController remote interface.
 * @see		GameController
 * @see		Game
 */
public class GameControllerImpl extends UnicastRemoteObject implements GameController
{
	/**
	 * Creates GameController object. Also creates a vector to hold the list of live games.
	 * @exeception	RemoteException
	 */
	public GameControllerImpl() throws RemoteException
	{
		games=new Vector(10,1);
		try {
			socketGate=new SocketGate();
		} catch (Exception e) {
			System.out.println("GameControllerImpl " + e);
		}
	}

	/**
	 * Changes the status of the argument game to 'request' and announces game to all players.
	 * @param	game the Game
	 * @exeception	RemoteException
	 */
	public synchronized boolean requestGame(Game game) throws RemoteException
	{
		if (noOfGames>=maxGames) {
			return false;
		} else {
			game.setStatus(Game.REQUEST);
			announceGame(game);
			return true;
		}
	}

	/**
	 * Announces that a game has been accepted to all players. Changes the game status to 'accept'
	 * before anouncement and adds it to the active games list, games.
	 * @param	game the Game
	 * @exeception	RemoteException
	 */
	public synchronized void acceptGame(Game game) throws RemoteException
	{
		++noOfGames;
		games.addElement(game);
		game.setStatus(Game.ACCEPT);
		announceGame(game);
	}

	/**
	 * Annouces to all players that a game has been declined. Changes the game status to 'decline'
	 * before announcement.
	 * @param	game the Game
	 * @exeception	RemoteException
	 */
	public synchronized void declineGame(Game game) throws RemoteException
	{
		game.setStatus(Game.DECLINE);
		announceGame(game);
	}

	/**
	 * Announces that a game has ended to all players. Changes the game status to 'end'
	 * before anouncement and removes it from the active games list, games.
	 * @param	game the Game
	 * @exeception	RemoteException
	 */
	public synchronized void endGame(Game game) throws RemoteException
	{
		--noOfGames;
		int index=findGamePosition(game);
		if (index!=-1) {
			games.removeElementAt(index);
			game.setStatus(Game.END);
			announceGame(game);
		}
	}

	/**
	 * Announces that a game has been won, drawn or lost to all players. Changes the game status to
	 * 'finish' before anouncement.
	 * @param	game the Game
	 * @exeception	RemoteException
	 */
	public synchronized void finishGame(Game game) throws RemoteException
	{
		--noOfGames;
		game.setStatus(Game.FINISHED);
		announceGame(game);
	}

	/**
	 * Announces that a player is busy and was not able to respond to a request. Changes the game
	 * status to 'busy' before anouncement.
	 * @param	game the Game
	 * @exeception	RemoteException
	 */
	public synchronized void playerIsBusy(Game game) throws RemoteException
	{
		game.setStatus(Game.BUSY);
		announceGame(game);
	}

	/**
	 * Mutator method to set the maximum allowable number of alive games.
	 * @param	max the maximum number of games.
	 */
	public synchronized void setMaxGames(int max)
	{
		maxGames=max;
	}

	private synchronized void announceGame(Game game)
	{
		System.out.println("Anouncement");
		socketGate.sendGame(game);
	}

	private synchronized int findGamePosition(Game game)
	{
		int i;
		int player=game.getPlayer().getID();
		int opponent=game.getOpponent().getID();
		int size=games.size();
		boolean found=false;
		for (i=0;i<size;i++) {
			if (DEBUG) System.out.println("looking at element " + i + " of " + size);
			Game g=(Game)games.elementAt(i);
			if (g.getPlayer().getID()==player && g.getOpponent().getID()==opponent) {
				if (DEBUG) System.out.println("found at " + i);
				found=true;
				return i;
			}
			if (g.getOpponent().getID()==player && g.getPlayer().getID()==opponent) {
				if (DEBUG) System.out.println("found at " + i);
				found=true;
				return i;
			}
		}
		if (DEBUG && found) System.out.println("Found game position at element " + i);
		return -1;
	}

	private int maxGames=0;
	private int noOfGames=0;
	private boolean DEBUG=false;
	private Vector games;
	private SocketGate socketGate;
}
