package c4.server;

import javax.swing.table.*;
import java.util.Vector;
import c4.shared.*;
import c4.shared.*;

/**
 * UI Model for ServGamesPane. Manages the display of the active games displayed in the ServGamesPane.
 * Extends the AbstractTableModel.
 * @see		ServGamesPane
 * @see		AbstractTableModel
 */
public class ServGamesModel extends AbstractTableModel
{
	protected static int NUM_COLUMNS=6;
	protected static int START_NUM_ROWS=4;
	protected int nextEmptyRow=0;
	protected int numRows=0;

	private boolean DEBUG=false;

	/** Column name.*/
	static final public String playerIDName="ID";
	/** Column name.*/
	static final public String playerName="Name";
	/** Column name.*/
	static final public String playerLocationName="Location";
	/** Column name.*/
	static final public String opponentIDName="ID";
	/** Column name.*/
	static final public String opponentName="Name";
	/** Column name.*/
	static final public String opponentLocationName="Location";

	protected Vector data=null;

	/**
	 * Creates the ServGamesModel object. Creates a vector to store the list of active games.
	 */
	public ServGamesModel()
	{
        	data=new Vector();
   	}

	/**
	 * Accessor method implemented from AbstractTableModel, returns the column names.
	 * @return	the column name.
	 * @see AbstractTableModel
	 */
   	public String getColumnName(int column)
   	{
		switch (column) {
		  	case 0:
		  		return playerIDName;
		  	case 1:
		    		return playerName;
		  	case 2:
		    		return playerLocationName;
		  	case 3:
		    		return opponentIDName;
	    	case 4:
		    		return opponentName;
		  	case 5:
	    		return opponentLocationName;
		}
		return "";
   	}


	/**
	 * Accessor method implemented from AbstractTableModel, returns the number of columns.
	 * @return	the number of columns.
	 * @see AbstractTableModel
	 */
    public int getColumnCount() {
       	return NUM_COLUMNS;
   	}

	/**
	 * Accessor method implemented from AbstractTableModel, returns the number of rows.
	 * @return	the number of rows.
	 * @see AbstractTableModel
	 */
    public int getRowCount()
    {
        if (numRows<START_NUM_ROWS) {
           		return START_NUM_ROWS;
        } else {
           		return numRows;
        }
    }

	/**
	 * Accessor method implemented from AbstractTableModel, returns the object by a specified row and
	 * column.
	 * @return	the object at a specified row and column.
	 * @see AbstractTableModel
	 */
    public Object getValueAt(int row,int column)
    {
		try {
	    	Game g=(Game)data.elementAt(row);
	    	switch (column) {
	    		case 0:
	    			return new Integer(g.getPlayer().getID());
	           	case 1:
					return g.getPlayer().getName();
				case 2:
					return g.getPlayer().getLocation();
				case 3:
					return new Integer(g.getOpponent().getID());
				case 4:
					return g.getOpponent().getName();
				case 5:
					return g.getOpponent().getLocation();
			}
		} catch (Exception e) {
		}
		return "";
	}

	/**
	 * Receives game objects which are then added or removed from the model.
	 * @param	game a game object to be removed or added in accordance with the game udpate type.
	 * @see		Game
	 */
   	public synchronized void updateGame(Game game)
   	{
   		int mode=game.getStatus();
   		if (DEBUG) System.out.println("Inside updateGame, update=" + mode);
   		if (mode==Game.ACCEPT) addGame(game);
   		else if (mode==Game.END || mode==Game.FINISHED) removeGame(game);
   	}

	/**
	 * Adds a game to the model.
	 * @param	game the game to be added.
	 */
   	public void addGame(Game game)
   	{
       	int player=game.getPlayer().getID();
       	int opponent=game.getOpponent().getID();
       	Game g=null;
       	int index=-1;
       	boolean found=false;
		boolean addedRow=false;

       	int i=0;
       	while (!found && (i<nextEmptyRow)) {
       		g=(Game)data.elementAt(i);
			if (g.getPlayer().getID()==player && g.getOpponent().getID()==opponent) {
           		found=true;
                index=i;
			} else if (g.getOpponent().getID()==player && g.getPlayer().getID()==opponent) {
                found=true;
               	index=i;
            } else {
               	i++;
            }
        }

        if (found) { //update old player
	    	data.setElementAt(game,index);
	    } else { //add new player
	   		if (numRows<=nextEmptyRow) {
			 	numRows++;
				addedRow=true;
           	}
 			index=nextEmptyRow;
	  		data.addElement(game);
		}
       	nextEmptyRow++;
		//Notify listeners that the data changed.
		if (addedRow) {
	    		fireTableRowsInserted(index,index);
			fireTableRowsUpdated(index,index);
		} else {
	    		fireTableRowsUpdated(index,index);
		}
	}

	/**
	 * Removes a game from the model.
	 * @param	game the game to be removed.
	 */
    	public void removeGame(Game game)
    	{
        	int player=game.getPlayer().getID();
        	int opponent=game.getOpponent().getID();
        	Game g=null;
        	int index=-1;
        	boolean found=false;
        	int i=0;
        	while (!found && (i<nextEmptyRow)) {
            		g=(Game)data.elementAt(i);
			if (g.getPlayer().getID()==player && g.getOpponent().getID()==opponent) {
                found=true;
                index=i;
			} else if (g.getOpponent().getID()==player && g.getPlayer().getID()==opponent) {
                found=true;
				index=i;
			} else {
				i++;
       		}
       	}

       	if (found) {
	   		data.remove(index);
			numRows--;
	   		fireTableRowsDeleted(index,index);
			fireTableRowsUpdated(index,nextEmptyRow);
           	index=--nextEmptyRow;
		}
    }

	/**
	 * Clears the model.
	 */
	public synchronized void clear()
    {
		int oldNumRows=numRows;
       	numRows=START_NUM_ROWS;
		data.removeAllElements();
       	nextEmptyRow=0;
		if (oldNumRows>START_NUM_ROWS) {
	   		fireTableRowsDeleted(START_NUM_ROWS,oldNumRows-1);
		}
		fireTableRowsUpdated(0,START_NUM_ROWS-1);
    }
}
